"""
Google Big Query connector is used to connect to dataset from
google big query api.
"""

from typing import Union

from sqlalchemy import create_engine

from pandasai.exceptions import InvalidConfigError

from ...connectors.base import BaseConnectorConfig
from ...connectors.sql import SQLBaseConnectorConfig, SQLConnector


class GoogleBigQueryConnectorConfig(SQLBaseConnectorConfig):
    """
    Connector configuration for big query.
    """

    credentials_path: str = None
    credentials_base64: str = None
    database: str
    table: str
    projectID: str


class GoogleBigQueryConnector(SQLConnector):
    """
    GoogleBigQuery Connectors are used to connect to BigQuery Data Cloud.
    """

    def __init__(self, config: Union[GoogleBigQueryConnectorConfig, dict]):
        """
        Initialize the GoogleBigQuery connector with the given configuration.

        Args:
            config (ConnectorConfig): The config for the GoogleBigQuery connector.
        """
        config["dialect"] = "bigquery"
        if isinstance(config, dict):
            env_vars = {
                "database": "BIG_QUERY_DATABASE",
                "credentials_path": "KEYFILE_PATH",
                "projectID": "PROJECT_ID",
            }
            config = self._populate_config_from_env(config, env_vars)

        if "credentials_base64" not in config and "credentials_path" not in config:
            raise InvalidConfigError(
                "credentials_path or credentials_base64 is needed to connect"
            )

        super().__init__(config)

    def _load_connector_config(self, config: Union[BaseConnectorConfig, dict]):
        return GoogleBigQueryConnectorConfig(**config)

    def _init_connection(self, config: GoogleBigQueryConnectorConfig):
        """
        Initialize Database Connection

        Args:
            config (GoogleBigQueryConnectorConfig): Configurations to load database

        """
        if config.credentials_path:
            self._engine = create_engine(
                f"{config.dialect}://{config.projectID}/{config.database}",
                credentials_path=config.credentials_path,
            )
        else:
            self._engine = create_engine(
                f"{config.dialect}://{config.projectID}/{config.database}?credentials_base64={config.credentials_base64}"
            )

        self._connection = self._engine.connect()

    def __repr__(self):
        """
        Return the string representation of the Google big query connector.

        Returns:
        str: The string representation of the Google big query connector.
        """
        return (
            f"<{self.__class__.__name__} dialect={self.config.dialect} "
            f"projectid= {self.config.projectID} database={self.config.database} >"
        )

    def equals(self, other):
        if isinstance(other, self.__class__):
            return (
                self.config.dialect,
                self.config.driver,
                self.config.credentials_path,
                self.config.credentials_base64,
                self.config.database,
                self.config.projectID,
            ) == (
                other.config.dialect,
                other.config.driver,
                other.config.credentials_path,
                other.config.credentials_base64,
                other.config.database,
                other.config.projectID,
            )
        return False



import re
from typing import Any, Iterable

import numpy as np

import pandasai.pandas as pd
from pandasai.exceptions import InvalidOutputValueMismatch


class OutputValidator:
    @staticmethod
    def validate(expected_type, result: dict[str, Any]) -> tuple[bool, Iterable[str]]:
        """
        Validate 'type' and 'value' from the result dict.

        Args:
            result (dict[str, Any]): The result of code execution in
                dict representation. Should have the following schema:
                {
                    "type": <expected_type_name>,
                    "value": <generated_value>
                }

        Returns:
             (tuple(bool, Iterable(str)):
                Boolean value whether the result matches output type
                and collection of logs containing messages about
                'type' or 'value' mismatches.
        """
        validation_logs = []
        result_type, result_value = result.get("type"), result.get("value")

        type_ok = OutputValidator.validate_type(result_type, expected_type)
        if not type_ok:
            validation_logs.append(
                f"The result dict contains inappropriate 'type'. "
                f"Expected '{expected_type}', actual '{result_type}'."
            )
        value_ok = OutputValidator.validate_value(result_value, expected_type)
        if not value_ok:
            validation_logs.append(
                f"result value {repr(result_value)} seems to be inappropriate "
                f"for the type '{expected_type}'."
            )

        return all((type_ok, value_ok)), validation_logs

    def validate_type(self, expected_type: str) -> bool:
        return self == expected_type if expected_type else True

    def validate_value(self, expected_type: str) -> bool:
        if not expected_type:
            return True
        elif expected_type == "number":
            return isinstance(self, (int, float))
        elif expected_type == "string":
            return isinstance(self, str)
        elif expected_type == "dataframe":
            return isinstance(self, (pd.DataFrame, pd.Series))
        elif expected_type == "plot":
            if not isinstance(self, (str, dict)):
                return False

            if isinstance(self, dict):
                return True

            path_to_plot_pattern = r"^(\/[\w.-]+)+(/[\w.-]+)*$|^[^\s/]+(/[\w.-]+)*$"
            return bool(re.match(path_to_plot_pattern, self))

    @staticmethod
    def validate_result(result: dict) -> bool:
        if not isinstance(result, dict) or "type" not in result:
            raise InvalidOutputValueMismatch(
                "Result must be in the format of dictionary of type and value"
            )

        if not result["type"]:
            return False

        elif result["type"] == "number":
            return isinstance(result["value"], (int, float, np.int64))
        elif result["type"] == "string":
            return isinstance(result["value"], str)
        elif result["type"] == "dataframe":
            return isinstance(result["value"], (pd.DataFrame, pd.Series))
        elif result["type"] == "plot":
            if "plotly" in repr(type(result["value"])):
                return True

            if not isinstance(result["value"], (str, dict)):
                return False

            if isinstance(result["value"], dict) or (
                isinstance(result["value"], str)
                and "data:image/png;base64" in result["value"]
            ):
                return True

            path_to_plot_pattern = r"^(\/[\w.-]+)+(/[\w.-]+)*$|^[^\s/]+(/[\w.-]+)*$"
            return bool(re.match(path_to_plot_pattern, result["value"]))



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Dict, Optional

from ..exceptions import APIKeyNotFoundError, UnsupportedModelError
from ..helpers import load_dotenv
from ..prompts.base import BasePrompt
from .base import LLM

if TYPE_CHECKING:
    from pandasai.pipelines.pipeline_context import PipelineContext


load_dotenv()


class BedrockClaude(LLM):
    """Bedrock Claude LLM
       Generates text using Anthropic Claude Messages API.

    Attributes:
        bedrock_runtime_client: The boto3 bedrock runtime client.
        max_tokens: Max number of tokens to generate.
        model: The Bedrock Claude model to use.
        temperature: (Optional) The amount of randomness injected into the response.
        top_p: (Optional) Use nucleus sampling. In nucleus sampling, Anthropic Claude computes the cumulative distribution over all the options for each subsequent token in decreasing probability order and cuts it off once it reaches a particular probability specified by top_p. You should alter either temperature or top_p, but not both.
        top_k: (Optional) Only sample from the top K options for each subsequent token.
        stop_sequences: (Optional) Custom text sequences that cause the model to stop generating. Anthropic Claude models normally stop when they have naturally completed their turn, in this case the value of the stop_reason response field is end_turn. If you want the model to stop generating when it encounters custom strings of text, you can use the stop_sequences parameter. If the model encounters one of the custom text strings, the value of the stop_reason response field is stop_sequence and the value of stop_sequence contains the matched stop sequence.
    """

    _supported__models = [
        "anthropic.claude-3-opus-20240229-v1:0",
        "anthropic.claude-3-5-sonnet-20240620-v1:0",
        "anthropic.claude-3-sonnet-20240229-v1:0",
        "anthropic.claude-3-haiku-20240307-v1:0",
    ]
    _valid_params = [
        "max_tokens",
        "model",
        "temperature",
        "top_p",
        "top_k",
        "stop_sequences",
    ]

    max_tokens: int = 1024
    model: str = "anthropic.claude-3-sonnet-20240229-v1:0"
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    top_k: Optional[float] = None
    stop_sequences: Optional[str] = None
    client: Any

    def __init__(self, bedrock_runtime_client, **kwargs):
        for key, val in kwargs.items():
            if key in self._valid_params:
                setattr(self, key, val)

        self.client = bedrock_runtime_client

        if self.model not in self._supported__models:
            raise UnsupportedModelError(self.model)

        invoke_model = getattr(self.client, "invoke_model", None)
        if not callable(invoke_model):
            raise APIKeyNotFoundError

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling text generation inference API."""
        return {
            "max_tokens": self.max_tokens,
            "model": self.model,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "stop_sequences": self.stop_sequences,
        }

    def call(self, instruction: BasePrompt, context: PipelineContext = None) -> str:
        prompt = instruction.to_string()
        memory = context.memory if context else None

        messages = []
        system_prompt = ""
        if memory:
            if memory.agent_info:
                system_prompt = memory.get_system_prompt()

            for message in memory.all():
                if message["is_user"]:
                    if messages and messages[-1]["role"] == "user":
                        messages[-1]["content"].append(
                            {"type": "text", "text": message["message"]}
                        )
                    else:
                        messages.append(
                            {
                                "role": "user",
                                "content": [
                                    {"type": "text", "text": message["message"]}
                                ],
                            }
                        )
                else:
                    messages.append(
                        {
                            "role": "assistant",
                            "content": [{"type": "text", "text": message["message"]}],
                        }
                    )

        # adding current prompt as latest query message
        if messages and messages[-1]["role"] == "user":
            messages[-1]["content"].append({"type": "text", "text": prompt})
        else:
            messages.append(
                {"role": "user", "content": [{"type": "text", "text": prompt}]},
            )

        params = {
            "anthropic_version": "bedrock-2023-05-31",
            "system": system_prompt,
            "messages": messages,
        }
        for key, value in self._default_params.items():
            if key != "model" and value is not None:
                params[key] = value

        body = json.dumps(params)
        # print(body)

        response = self.client.invoke_model(modelId=self.model, body=body)

        response_body = json.loads(response.get("body").read())

        self.last_prompt = prompt
        # print(response_body["content"][0]["text"])
        return response_body["content"][0]["text"]

    @property
    def type(self) -> str:
        return "bedrock-claude"



import json
from enum import Enum

import yaml

import pandasai.pandas as pd


class DataframeSerializerType(Enum):
    JSON = 1
    YML = 2
    CSV = 3
    SQL = 4


class DataframeSerializer:
    def __init__(self) -> None:
        pass

    def serialize(
        self,
        df: pd.DataFrame,
        extras: dict = None,
        type_: DataframeSerializerType = DataframeSerializerType.YML,
    ) -> str:
        if type_ == DataframeSerializerType.YML:
            return self.convert_df_to_yml(df, extras)
        elif type_ == DataframeSerializerType.JSON:
            return self.convert_df_to_json_str(df, extras)
        elif type_ == DataframeSerializerType.SQL:
            return self.convert_df_sql_connector_to_str(df, extras)
        else:
            return self.convert_df_to_csv(df, extras)

    def convert_df_to_csv(self, df: pd.DataFrame, extras: dict) -> str:
        """
        Convert df to csv like format where csv is wrapped inside <dataframe></dataframe>
        Args:
            df (pd.DataFrame): PandasAI dataframe or dataframe
            extras (dict, optional): expect index to exists

        Returns:
            str: dataframe stringify
        """
        dataframe_info = "<dataframe"

        # Add name attribute if available
        if df.name is not None:
            dataframe_info += f' name="{df.name}"'

        # Add description attribute if available
        if df.description is not None:
            dataframe_info += f' description="{df.description}"'

        dataframe_info += ">"

        # Add dataframe details
        dataframe_info += f"\ndfs[{extras['index']}]:{df.rows_count}x{df.columns_count}\n{df.to_csv()}"

        # Close the dataframe tag
        dataframe_info += "</dataframe>\n"

        return dataframe_info

    def convert_df_sql_connector_to_str(
        self, df: pd.DataFrame, extras: dict = None
    ) -> str:
        """
        Convert df to csv like format where csv is wrapped inside <table></table>
        Args:
            df (pd.DataFrame): PandasAI dataframe or dataframe
            extras (dict, optional): expect index to exists

        Returns:
            str: dataframe stringify
        """
        table_description_tag = (
            f' description="{df.description}"' if df.description is not None else ""
        )
        table_head_tag = f'<table name="{df.name}"{table_description_tag}>'
        return f"{table_head_tag}\n{df.get_head().to_csv()}\n</table>"

    def convert_df_to_json(self, df: pd.DataFrame, extras: dict) -> dict:
        """
        Convert df to json dictionary and return json
        Args:
            df (pd.DataFrame): PandasAI dataframe or dataframe
            extras (dict, optional): expect index to exists

        Returns:
            str: dataframe json
        """
        # Initialize the result dictionary
        df_number_key = f"dfs[{extras['index']}]"

        # Create a dictionary representing the data structure
        df_info = {
            "name": df.name,
            "description": df.description,
            "type": (
                df.type
                if "is_direct_sql" in extras and extras["is_direct_sql"]
                else extras["type"]
            ),
        }
        # Add DataFrame details to the result
        data = {
            "rows": df.rows_count,
            "columns": df.columns_count,
            "schema": {"fields": []},
        }

        # Iterate over DataFrame columns
        df_head = df.get_head()
        for col_name, col_dtype in df_head.dtypes.items():
            col_info = {
                "name": col_name,
                "type": str(col_dtype),
            }

            if not extras.get("enforce_privacy") or df.custom_head is not None:
                col_info["samples"] = df_head[col_name].head().tolist()

            # Add column description if available
            if df.field_descriptions and isinstance(df.field_descriptions, dict):
                if col_description := df.field_descriptions.get(col_name, None):
                    col_info["description"] = col_description

            if df.connector_relations:
                for relation in df.connector_relations:
                    from pandasai.ee.connectors.relations import ForeignKey, PrimaryKey

                    if (
                        isinstance(relation, PrimaryKey) and relation.name == col_name
                    ) or (
                        isinstance(relation, ForeignKey) and relation.field == col_name
                    ):
                        col_info["constraints"] = relation.to_string()

            data["schema"]["fields"].append(col_info)

        result = df_info | data

        if "is_direct_sql" in extras and extras["is_direct_sql"]:
            return result

        return {df_number_key: result}

    def convert_df_to_json_str(self, df: pd.DataFrame, extras: dict) -> str:
        """
        Convert df to json and return it as string
        Args:
            df (pd.DataFrame): PandasAI dataframe or dataframe
            extras (dict, optional): expect index to exists

        Returns:
            str: dataframe stringify
        """
        return json.dumps(self.convert_df_to_json(df, extras))

    def convert_df_to_yml(self, df: pd.DataFrame, extras: dict) -> str:
        json_df = self.convert_df_to_json(df, extras)
        yml_str = yaml.dump(json_df, sort_keys=False)
        if "is_direct_sql" in extras and extras["is_direct_sql"]:
            return f"<table>\n{yml_str}\n</table>\n"
        return yml_str



import uuid
from typing import Callable, Iterable, List, Optional, Union

import pinecone

from pandasai.helpers.logger import Logger
from pandasai.vectorstores.vectorstore import VectorStore


class Pinecone(VectorStore):
    """
    Implementation of Pinecone vector store
    """

    _logger: Logger

    def __init__(
        self,
        api_key: str,
        index: Union[str, pinecone.Index] = "pandasai",
        embedding_function: Optional[Callable[[List[str]], List[float]]] = None,
        dimensions=1536,
        metric="cosine",
        pool_threads: int = 1,
        specs: pinecone.ServerlessSpec = None,
        max_samples: int = 1,
        similary_threshold: int = 1.5,
        logger: Optional[Logger] = None,
    ) -> None:
        self._logger = logger or Logger()
        self._max_samples = max_samples
        self._similarity_threshold = similary_threshold
        self._api_key = api_key

        self._metatext_key = "text"

        self._embedding_function = embedding_function

        self._pinecone = pinecone.Pinecone(api_key=api_key, pool_threads=pool_threads)

        if isinstance(index, str):
            if index not in self._pinecone.list_indexes().names():
                self._index = self._pinecone.create_index(
                    name=index,
                    dimension=dimensions,
                    metric=metric,
                    spec=specs
                    or pinecone.ServerlessSpec(cloud="aws", region="us-east-1"),
                )

            self._index = self._pinecone.Index(name=index)

        else:
            self._index = index

        self._logger.log("Successfully initialized index")

    def add_question_answer(
        self,
        queries: Iterable[str],
        codes: Iterable[str],
        ids: Optional[Iterable[str]] = None,
        metadatas: Optional[List[dict]] = None,
    ) -> List[str]:
        """
        Add question and answer(code) to the training set
        Args:
            query: string of question
            code: str
            ids: Optional Iterable of ids associated with the texts.
            metadatas: Optional list of metadatas associated with the texts.
            kwargs: vectorstore specific parameters
        Returns:
            List of ids from adding the texts into the vectorstore.
        """
        if len(queries) != len(codes):
            raise ValueError(
                f"Queries and codes dimension doesn't match {len(queries)} != {len(codes)}"
            )

        if ids is None:
            ids = [f"{str(uuid.uuid4())}-qa" for _ in queries]

        metadatas = metadatas or [{} for _ in ids]

        qa_str = [self._format_qa(query, code) for query, code in zip(queries, codes)]

        for index, metadata in enumerate(metadatas):
            metadata[self._metatext_key] = qa_str[index]

        vector_data = [
            {"id": ids[index], "values": qa, "metadata": metadatas[index]}
            for index, qa in enumerate(self._embedding_function(qa_str))
        ]

        self._index.upsert(vectors=vector_data, namespace="qa")

        return ids

    def add_docs(
        self,
        docs: Iterable[str],
        ids: Optional[Iterable[str]] = None,
        metadatas: Optional[List[dict]] = None,
    ) -> List[str]:
        """
        Add docs to the training set
        Args:
            docs: Iterable of strings to add to the vectorstore.
            ids: Optional Iterable of ids associated with the texts.
            metadatas: Optional list of metadatas associated with the texts.
            kwargs: vectorstore specific parameters

        Returns:
            List of ids from adding the texts into the vectorstore.
        """
        if not isinstance(docs, list):
            raise ValueError("Docs must be list of strings!")

        if ids is None:
            ids = [f"{str(uuid.uuid4())}-docs" for _ in docs]

        metadatas = metadatas or [{} for _ in ids]

        doc_embeddings = self._embedding_function(docs)

        for index, metadata in enumerate(metadatas):
            metadata[self._metatext_key] = docs[index]

        vector_data = [
            {"id": ids[index], "values": doc, "metadata": metadatas[index]}
            for index, doc in enumerate(doc_embeddings)
        ]

        self._index.upsert(vectors=vector_data, namespace="docs")

        return ids

    def update_question_answer(
        self,
        ids: Iterable[str],
        queries: Iterable[str],
        codes: Iterable[str],
        metadatas: Optional[List[dict]] = None,
    ) -> List[str]:
        """
        Update question and answer(code) to the training set
        Args:
            ids: Iterable of ids associated with the texts.
            queries: string of question
            codes: str
            metadatas: Optional list of metadatas associated with the texts.
            kwargs: vectorstore specific parameters
        Returns:
            List of ids from updating the texts into the vectorstore.
        """
        if len(queries) != len(codes):
            raise ValueError(
                f"Queries and codes dimension doesn't match {len(queries)} != {len(codes)}"
            )

        qa_str = [self._format_qa(query, code) for query, code in zip(queries, codes)]

        metadatas = metadatas or [{} for _ in ids]

        for index, metadata in enumerate(metadatas):
            metadata[self._metatext_key] = qa_str[index]

        for index, qa in enumerate(self._embedding_function(qa_str)):
            self._index.update(
                id=ids[index], values=qa, set_metadata=metadatas[index], namespace="qa"
            )

    def update_docs(
        self,
        ids: Iterable[str],
        docs: Iterable[str],
        metadatas: Optional[List[dict]] = None,
    ) -> List[str]:
        """
        Update docs to the training set
        Args:
            ids: Iterable of ids associated with the texts.
            docs: Iterable of strings to update to the vectorstore.
            metadatas: Optional list of metadatas associated with the texts.
            kwargs: vectorstore specific parameters

        Returns:
            List of ids from adding the texts into the vectorstore.
        """

        doc_embeddings = self._embedding_function(docs)

        metadatas = metadatas or [{} for _ in ids]

        for index, metadata in enumerate(metadatas):
            metadata[self._metatext_key] = docs[index]

        for index, doc in enumerate(doc_embeddings):
            self._index.update(
                id=ids[index],
                values=doc,
                set_metadata=metadatas[index],
                namespace="docs",
            )

    def delete_question_and_answers(
        self, ids: Optional[List[str]] = None
    ) -> Optional[bool]:
        """
        Delete by vector ID to delete question and answers
        Args:
            ids: List of ids to delete

        Returns:
            Optional[bool]: True if deletion is successful,
            False otherwise
        """

        self._index.delete(ids=ids, namespace="qa")
        return True

    def delete_docs(self, ids: Optional[List[str]] = None) -> Optional[bool]:
        """
        Delete by vector ID to delete docs
        Args:
            ids: List of ids to delete

        Returns:
            Optional[bool]: True if deletion is successful,
            False otherwise
        """
        self._index.delete(ids=ids, namespace="docs")
        return True

    def get_relevant_question_answers(
        self, question: str, k: Union[int, None] = None
    ) -> List[dict]:
        """
        Returns relevant question answers based on search
        """
        k = k or self._max_samples

        questions = self._embedding_function([question])

        results = self._index.query(
            vector=questions,
            top_k=k,
            include_metadata=True,
            namespace="qa",
            include_values=True,
        )

        return self._filter_docs_based_on_distance(results, self._similarity_threshold)

    def get_relevant_docs(self, question: str, k: int = None) -> List[dict]:
        """
        Returns relevant documents based search
        """
        k = k or self._max_samples

        questions = self._embedding_function([question])

        results = self._index.query(
            vector=questions,
            top_k=k,
            include_metadata=True,
            namespace="docs",
            include_values=True,
        )

        return self._filter_docs_based_on_distance(results, self._similarity_threshold)

    def get_relevant_question_answers_by_id(self, ids: Iterable[str]) -> List[dict]:
        """
        Returns relevant question answers based on ids
        """
        return self._index.fetch(id=ids, namespace="qa")

    def get_relevant_docs_by_id(self, ids: Iterable[str]) -> List[dict]:
        """
        Returns relevant question answers based on ids
        """

        return self._index.fetch(id=ids, namespace="docs")

    def get_relevant_qa_documents(self, question: str, k: int = None) -> List[str]:
        """
        Returns relevant question answers documents only
        Args:
            question (_type_): list of documents
        """
        return self.get_relevant_question_answers(question, k)["documents"][0]

    def get_relevant_docs_documents(self, question: str, k: int = None) -> List[str]:
        """
        Returns relevant question answers documents only
        Args:
            question (_type_): list of documents
        """
        return self.get_relevant_docs(question, k)["documents"][0]

    def _filter_docs_based_on_distance(self, documents, threshold: int) -> List[str]:
        """
        Filter documents based on threshold
        Args:
            documents (List[str]): list of documents in string
            distances (List[float]): list of distances in float
            threshold (int): similarity threshold

        Returns:
            _type_: _description_
        """
        filtered_data = [
            (
                document["metadata"][self._metatext_key],
                document["score"],
                document["metadata"],
                document["id"],
            )
            for document in documents["matches"]
            if document["score"] < threshold
        ]

        return {
            key: [[data[i] for data in filtered_data]]
            for i, key in enumerate(["documents", "distances", "metadata", "ids"])
        }



from __future__ import annotations

try:
    from langchain_core.language_models import BaseLanguageModel
    from langchain_core.language_models.chat_models import BaseChatModel

except ImportError:
    from unittest.mock import Mock

    # Fallback definitions if langchain_core is not installed
    BaseLanguageModel = BaseChatModel = Mock

from typing import TYPE_CHECKING

from pandasai.prompts.base import BasePrompt

from .base import LLM

if TYPE_CHECKING:
    from pandasai.pipelines.pipeline_context import PipelineContext

"""Langchain LLM 

This module is to run LLM using LangChain framework.

Example:
    Use below example to call LLM
    >>> from pandasai.llm.langchain import LangchainLLm
"""


def is_langchain_llm(llm) -> bool:
    return isinstance(llm, BaseLanguageModel)


class LangchainLLM(LLM):
    """
    Class to wrap Langchain LLMs and make PandasAI interoperable
    with LangChain.
    """

    langchain_llm: BaseLanguageModel

    def __init__(self, langchain_llm: BaseLanguageModel):
        self.langchain_llm = langchain_llm

    def call(
        self, instruction: BasePrompt, context: PipelineContext = None, suffix: str = ""
    ) -> str:
        prompt = instruction.to_string() + suffix
        memory = context.memory if context else None
        prompt = self.prepend_system_prompt(prompt, memory)
        self.last_prompt = prompt

        res = self.langchain_llm.invoke(prompt)
        return res.content if isinstance(self.langchain_llm, BaseChatModel) else res

    @property
    def type(self) -> str:
        return f"langchain_{self.langchain_llm._llm_type}"



from typing import Optional, Union

from pandasai.agent.base_judge import BaseJudge
from pandasai.config import load_config_from_json
from pandasai.ee.agents.judge_agent.pipeline.judge_pipeline import JudgePipeline
from pandasai.pipelines.abstract_pipeline import AbstractPipeline
from pandasai.pipelines.judge.judge_pipeline_input import JudgePipelineInput
from pandasai.pipelines.pipeline_context import PipelineContext
from pandasai.schemas.df_config import Config


class JudgeAgent(BaseJudge):
    def __init__(
        self,
        config: Optional[Union[Config, dict]] = None,
        pipeline: AbstractPipeline = None,
    ) -> None:
        context = None
        if config:
            if isinstance(config, dict):
                config = Config(**load_config_from_json(config))

            context = PipelineContext(None, config)

        pipeline = pipeline or JudgePipeline(context=context)
        super().__init__(pipeline)

    def evaluate(self, query: str, code: str) -> bool:
        input_data = JudgePipelineInput(query, code)
        return self.pipeline.run(input_data)



import glob
import os
from typing import Any

import duckdb

from ..constants import CACHE_TOKEN, DEFAULT_FILE_PERMISSIONS
from .path import find_project_root


class Cache:
    """Cache class for caching queries. It is used to cache queries
    to save time and money.

    Args:
        filename (str): filename to store the cache.
    """

    def __init__(self, filename="cache_db_0.11", abs_path=None):
        # Define cache directory and create directory if it does not exist
        if abs_path:
            cache_dir = abs_path
        else:
            try:
                cache_dir = os.path.join(find_project_root(), "cache")
            except ValueError:
                cache_dir = os.path.join(os.getcwd(), "cache")

        os.makedirs(cache_dir, mode=DEFAULT_FILE_PERMISSIONS, exist_ok=True)

        self.filepath = os.path.join(cache_dir, f"{filename}.db")
        self.connection = duckdb.connect(self.filepath)
        self.connection.execute(
            "CREATE TABLE IF NOT EXISTS cache (key STRING, value STRING)"
        )

    def versioned_key(self, key: str) -> str:
        return f"{CACHE_TOKEN}-{key}"

    def set(self, key: str, value: str) -> None:
        """Set a key value pair in the cache.

        Args:
            key (str): key to store the value.
            value (str): value to store in the cache.
        """
        self.connection.execute(
            "INSERT INTO cache VALUES (?, ?)", [self.versioned_key(key), value]
        )

    def get(self, key: str) -> str:
        """Get a value from the cache.

        Args:
            key (str): key to get the value from the cache.

        Returns:
            str: value from the cache.
        """
        result = self.connection.execute(
            "SELECT value FROM cache WHERE key=?", [self.versioned_key(key)]
        )
        return row[0] if (row := result.fetchone()) else None

    def delete(self, key: str) -> None:
        """Delete a key value pair from the cache.

        Args:
            key (str): key to delete the value from the cache.
        """
        self.connection.execute(
            "DELETE FROM cache WHERE key=?", [self.versioned_key(key)]
        )

    def close(self) -> None:
        """Close the cache."""
        self.connection.close()

    def clear(self) -> None:
        """Clean the cache."""
        self.connection.execute("DELETE FROM cache")

    def destroy(self) -> None:
        """Destroy the cache."""
        self.connection.close()
        for cache_file in glob.glob(f"{self.filepath}.*"):
            os.remove(cache_file)

    def get_cache_key(self, context: Any) -> str:
        """
        Return the cache key for the current conversation.

        Returns:
            str: The cache key for the current conversation
        """
        cache_key = context.memory.get_conversation()

        # make the cache key unique for each combination of dfs
        for df in context.dfs:
            cache_key += str(df.column_hash)

        return cache_key



from abc import ABC, abstractmethod
from typing import Any

from pandasai.pipelines.logic_unit_output import LogicUnitOutput


class BaseLogicUnit(ABC):
    """
    Logic units for pipeline each logic unit should be inherited from this Logic unit
    """

    def __init__(self, skip_if=None, on_execution=None, before_execution=None):
        super().__init__()
        self.skip_if = skip_if
        self.on_execution = on_execution
        self.before_execution = before_execution

    @abstractmethod
    def execute(self, input: Any, **kwargs) -> LogicUnitOutput:
        """
        This method will return output according to
        Implementation.

        :param input: Your input data.
        :param kwargs: A dictionary of keyword arguments.
            - 'logger' (any): The logger for logging.
            - 'config' (Config): Global configurations for the test
            - 'context' (any): The execution context.

        :return: The result of the execution.
        """
        raise NotImplementedError("execute method is not implemented.")



import pandasai.pandas as pd

from .from_google_sheets import from_google_sheets


class FileImporter:
    """
    Class to import a dataframe from a file (csv, parquet, xlsx)
    """

    @staticmethod
    def import_from_file(file_path: str) -> pd.DataFrame:
        """
        Import a dataframe from a file (csv, parquet, xlsx)

        Args:
            file_path (str): Path to the file to be imported.

        Returns:
            pd.DataFrame: Pandas dataframe
        """

        if file_path.endswith(".csv"):
            return pd.read_csv(file_path)
        elif file_path.endswith(".parquet"):
            return pd.read_parquet(file_path)
        elif file_path.endswith(".xlsx"):
            return pd.read_excel(file_path)
        elif file_path.startswith("https://docs.google.com/spreadsheets/"):
            return from_google_sheets(file_path)[0]
        else:
            raise ValueError("Invalid file format.")



import os
import platform

import requests

from pandasai.__version__ import __version__


def scarf_analytics():
    try:
        if (
            os.getenv("SCARF_NO_ANALYTICS") != "true"
            and os.getenv("DO_NOT_TRACK") != "true"
        ):
            requests.get(
                "https://package.pandabi.ai/pandasai-telemetry?version="
                + __version__
                + "&platform="
                + platform.system()
            )
    except Exception:
        pass



"""IBM watsonx

This module is for the IBM watsonx.ai large language models.

"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING, Optional

from ..exceptions import APIKeyNotFoundError
from ..helpers import load_dotenv
from ..helpers.optional import import_dependency
from ..prompts.base import BasePrompt
from .base import LLM

if TYPE_CHECKING:
    from pandasai.pipelines.pipeline_context import PipelineContext


load_dotenv()


class IBMwatsonx(LLM):
    decoding_method: Optional[str] = None
    length_penalty: Optional[dict] = None
    repetition_penalty: Optional[int] = None
    stop_sequences: Optional[list] = None
    truncate_input_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    top_k: Optional[int] = None
    min_new_tokens: Optional[int] = None
    max_new_tokens: Optional[int] = None
    random_seed: Optional[int] = None

    def __init__(
        self,
        model: Optional[str] = None,
        api_key: Optional[str] = None,
        watsonx_url: Optional[str] = None,
        watsonx_project_id: str = None,
        **kwargs,
    ):
        """
        __init__ method of IBMwatsonx class

        Args:
            api_key (str): watsonx API key
                 To determine apikey go to https://cloud.ibm.com/iam/apikeys
            watsonx_url (str): watsonx endpoint url
                Depends on which region the Watson Studio service was provisioned.
                You can find available urls here:
                https://ibm.github.io/watsonx-ai-python-sdk/setup_cloud.html#authentication
            project_id (str) : watsonx project id
                ID of the Watson Studio project
                You can copy the project_id from Project’s Manage tab (Project -> Manage -> General -> Details).
        """
        self.model = model or "ibm/granite-13b-chat-v2"

        self.api_key = api_key or os.getenv("WATSONX_API_KEY")
        self.project_id = watsonx_project_id or os.getenv("WATSONX_PROJECT_ID")
        self.watsonx_url = watsonx_url or os.getenv("WATSONX_URL")

        if self.api_key is None:
            raise APIKeyNotFoundError(
                "IBM watsonx API key is required. Please add an environment variable "
                "`WATSONX_API_KEY` or pass `api_key` as a named parameter. "
                "You can create one at https://cloud.ibm.com/iam/apikeys."
            )
        if self.watsonx_url is None:
            raise APIKeyNotFoundError(
                "IBM watsonx region url is required. Please add an environment variable "
                "`WATSONX_URL` or pass `watsonx_url` as a named parameter."
            )
        if self.project_id is None:
            raise APIKeyNotFoundError(
                "Project ID is required. Please add an environment variable "
                "`WATSONX_PROJECT_ID` or pass `project_id` as a named parameter."
                "The project ID can be found in the created Watson Studio resource in IBM Cloud."
            )
        self._set_params(**kwargs)
        self._configure(
            api_key=self.api_key, url=self.watsonx_url, project_id=self.project_id
        )

    def _configure(self, api_key: str, url: str, project_id: str):
        """
        Configure IBM watsonx.
        """
        from ibm_watsonx_ai.foundation_models import Model

        self.watsonx = Model(
            model_id=self.model,
            params={
                "decoding_method": self.decoding_method,
                "length_penalty": self.length_penalty,
                "repetition_penalty": self.repetition_penalty,
                "stop_sequences": self.stop_sequences,
                "temperature": self.temperature,
                "top_p": self.top_p,
                "top_k": self.top_k,
                "min_new_tokens": self.min_new_tokens,
                "max_new_tokens": self.max_new_tokens,
                "truncate_input_tokens": self.truncate_input_tokens,
                "random_seed": self.random_seed,
            },
            credentials={"apikey": api_key, "url": url},
            project_id=project_id,
        )

    def _set_params(self, **kwargs):
        """
        Dynamically set Parameters for the object.

        Args:
            **kwargs:
                Possible keyword arguments: "temperature", "top_p", "top_k",
                "max_output_tokens".

        Returns:
            None.

        """
        err_msg = "Install ibm-watsonx-ai for IBMwatsonx"
        import_dependency("ibm_watsonx_ai", extra=err_msg)

        from ibm_watsonx_ai.metanames import GenTextParamsMetaNames

        valid_params = GenTextParamsMetaNames().get_example_values().keys()
        for key, value in kwargs.items():
            if key in valid_params:
                setattr(self, key, value)
            else:
                raise KeyError(
                    f"Parameter {key} is invalid. Accepted parameters: {[*valid_params]}"
                )

    def call(self, instruction: BasePrompt, context: PipelineContext = None) -> str:
        prompt = instruction.to_string()

        memory = context.memory if context else None

        prompt = self.prepend_system_prompt(prompt, memory)

        self.last_prompt = prompt

        return self.watsonx.generate_text(prompt=prompt)

    @property
    def type(self) -> str:
        return "ibm-watsonx-ai"



import traceback
from typing import Any, Callable

from pandasai.exceptions import ExecuteSQLQueryNotUsed, InvalidLLMOutputType
from pandasai.helpers.logger import Logger
from pandasai.pipelines.base_logic_unit import BaseLogicUnit
from pandasai.pipelines.chat.error_correction_pipeline.error_correction_pipeline_input import (
    ErrorCorrectionPipelineInput,
)
from pandasai.pipelines.logic_unit_output import LogicUnitOutput
from pandasai.pipelines.pipeline_context import PipelineContext
from pandasai.prompts.base import BasePrompt
from pandasai.prompts.correct_error_prompt import CorrectErrorPrompt
from pandasai.prompts.correct_execute_sql_query_usage_error_prompt import (
    CorrectExecuteSQLQueryUsageErrorPrompt,
)
from pandasai.prompts.correct_output_type_error_prompt import (
    CorrectOutputTypeErrorPrompt,
)


class ErrorPromptGeneration(BaseLogicUnit):
    on_prompt_generation: Callable[[str], None]

    def __init__(
        self,
        on_prompt_generation=None,
        skip_if=None,
        on_execution=None,
        before_execution=None,
    ):
        self.on_prompt_generation = on_prompt_generation
        super().__init__(skip_if, on_execution, before_execution)

    def execute(self, input: ErrorCorrectionPipelineInput, **kwargs) -> Any:
        """
        A method to retry the code execution with error correction framework.

        Args:
            code (str): A python code
            context (PipelineContext) : Pipeline Context
            logger (Logger) : Logger
            e (Exception): An exception
            dataframes

        Returns (str): A python code
        """
        self.context: PipelineContext = kwargs.get("context")
        self.logger: Logger = kwargs.get("logger")
        e = input.exception

        prompt = self.get_prompt(e, input.code)
        if self.on_prompt_generation:
            self.on_prompt_generation(prompt)

        self.logger.log(f"Using prompt: {prompt}")

        return LogicUnitOutput(
            prompt,
            True,
            "Prompt Generated Successfully",
            {
                "content_type": "prompt",
                "value": prompt.to_string(),
            },
        )

    def get_prompt(self, e: Exception, code: str) -> BasePrompt:
        """
        Return a prompt by key.

        Args:
            values (dict): The values to use for the prompt

        Returns:
            BasePrompt: The prompt
        """
        traceback_errors = traceback.format_exc()
        return (
            CorrectOutputTypeErrorPrompt(
                context=self.context,
                code=code,
                error=traceback_errors,
                output_type=self.context.get("output_type"),
            )
            if isinstance(e, InvalidLLMOutputType)
            else (
                CorrectExecuteSQLQueryUsageErrorPrompt(
                    context=self.context, code=code, error=traceback_errors
                )
                if isinstance(e, ExecuteSQLQueryNotUsed)
                else CorrectErrorPrompt(
                    context=self.context,
                    code=code,
                    error=traceback_errors,
                )
            )
        )



"""
SQL connectors are used to connect to SQL databases in different dialects.
"""

import hashlib
import os
import re
import time
from functools import cache, cached_property
from typing import Optional, Union

import sqlglot
from sqlalchemy import asc, create_engine, select, text
from sqlalchemy.engine import Connection

import pandasai.pandas as pd
from pandasai.exceptions import MaliciousQueryError
from pandasai.helpers.path import find_project_root

from ..constants import DEFAULT_FILE_PERMISSIONS
from .base import BaseConnector, BaseConnectorConfig


class SQLBaseConnectorConfig(BaseConnectorConfig):
    """
    Base Connector configuration.
    """

    driver: Optional[str] = None
    dialect: Optional[str] = None


class SqliteConnectorConfig(SQLBaseConnectorConfig):
    """
    Connector configurations for sqlite db.
    """

    table: str
    database: str


class SQLConnectorConfig(SQLBaseConnectorConfig):
    """
    Connector configuration.
    """

    host: str
    port: int
    username: str
    password: str


class SQLConnector(BaseConnector):
    """
    SQL connectors are used to connect to SQL databases in different dialects.
    """

    _engine = None
    _connection: Connection = None
    _rows_count: int = None
    _columns_count: int = None
    _cache_interval: int = 600  # 10 minutes

    def __init__(
        self,
        config: Union[BaseConnectorConfig, dict],
        cache_interval: int = 600,
        **kwargs,
    ):
        """
        Initialize the SQL connector with the given configuration.

        Args:
            config (ConnectorConfig): The configuration for the SQL connector.
        """
        config = self._load_connector_config(config)
        super().__init__(config, **kwargs)

        if config.dialect is None:
            raise Exception("SQL dialect must be specified")

        self._init_connection(config)

        self._cache_interval = cache_interval

        # Table to equal to table name for sql connectors
        self.name = self.fallback_name

    def _load_connector_config(self, config: Union[BaseConnectorConfig, dict]):
        """
        Loads passed Configuration to object

        Args:
            config (BaseConnectorConfig): Construct config in structure

        Returns:
            config: BaseConenctorConfig
        """
        return SQLConnectorConfig(**config)

    def _init_connection(self, config: SQLConnectorConfig):
        """
        Initialize Database Connection

        Args:
            config (SQLConnectorConfig): Configurations to load database

        """

        if config.driver:
            self._engine = create_engine(
                f"{config.dialect}+{config.driver}://{config.username}:{config.password}"
                f"@{config.host}:{str(config.port)}/{config.database}",
                connect_args=config.connect_args,
            )
        else:
            self._engine = create_engine(
                f"{config.dialect}://{config.username}:{config.password}@{config.host}"
                f":{str(config.port)}/{config.database}",
                connect_args=config.connect_args,
            )

        self._connection = self._engine.connect()

    def __del__(self):
        """
        Close the connection to the SQL database.
        """
        if self._connection:
            self._connection.close()

    def __repr__(self):
        """
        Return the string representation of the SQL connector.

        Returns:
            str: The string representation of the SQL connector.
        """
        return (
            f"<{self.__class__.__name__} dialect={self.config.dialect} "
            f"driver={self.config.driver} host={self.config.host} "
            f"port={str(self.config.port)} database={self.config.database} "
            f"table={self.config.table}>"
        )

    def _validate_column_name(self, column_name):
        regex = r"^[a-zA-Z0-9_]+$"
        if not re.match(regex, column_name):
            raise ValueError(f"Invalid column name: {column_name}")

    def _build_query(self, limit=None, order=None):
        base_query = select("*").select_from(text(self.cs_table_name))
        if self.config.where or self._additional_filters:
            # conditions is the list of where + additional filters
            conditions = []
            if self.config.where:
                conditions += self.config.where
            if self._additional_filters:
                conditions += self._additional_filters

            query_params = {}
            condition_strings = []

            valid_operators = ["=", ">", "<", ">=", "<=", "LIKE", "!=", "IN", "NOT IN"]

            for i, condition in enumerate(conditions):
                if len(condition) == 3:
                    column_name, operator, value = condition
                    if operator in valid_operators:
                        self._validate_column_name(column_name)

                        condition_strings.append(f"{column_name} {operator} :value_{i}")
                        query_params[f"value_{i}"] = value

            if condition_strings:
                where_clause = " AND ".join(condition_strings)
                base_query = base_query.where(
                    text(where_clause).bindparams(**query_params)
                )

        if order:
            base_query = base_query.order_by(asc(text(order)))

        if limit:
            base_query = base_query.limit(limit)

        return base_query

    @cache
    def head(self, n: int = 5) -> pd.DataFrame:
        """
        Return the head of the data source that the connector is connected to.
        This information is passed to the LLM to provide the schema of the data source.

        Returns:
            DataFrame: The head of the data source.
        """

        if self.logger:
            self.logger.log(
                f"Getting head of {self.config.table} "
                f"using dialect {self.config.dialect}"
            )

        # Run a SQL query to get all the columns names and 5 random rows
        query = self._build_query(limit=n, order="RAND()")

        # Return the head of the data source
        return pd.read_sql(query, self._connection)

    def _get_cache_path(self, include_additional_filters: bool = False):
        """
        Return the path of the cache file.

        Args:
            include_additional_filters (bool, optional): Whether to include the
                additional filters in when calling `_get_column_hash()`.
                Defaults to False.

        Returns:
            str: The path of the cache file.
        """
        try:
            cache_dir = os.path.join((find_project_root()), "cache")
        except ValueError:
            cache_dir = os.path.join(os.getcwd(), "cache")

        os.makedirs(cache_dir, mode=DEFAULT_FILE_PERMISSIONS, exist_ok=True)

        filename = (
            self._get_column_hash(include_additional_filters=include_additional_filters)
            + ".parquet"
        )
        path = os.path.join(cache_dir, filename)

        return path

    def _cached(self, include_additional_filters: bool = False) -> Union[str, bool]:
        """
        Return the cached data if it exists and is not older than the cache interval.

        Args:
            include_additional_filters (bool, optional): Whether to include the
                additional filters in when calling `_get_column_hash()`.
                Defaults to False.

        Returns:
            DataFrame|bool: The name of the file containing cached data if it exists
                and is not older than the cache interval, False otherwise.
        """
        filename = self._get_cache_path(
            include_additional_filters=include_additional_filters
        )
        if not os.path.exists(filename):
            return False

        # If the file is older than 1 day, delete it
        if os.path.getmtime(filename) < time.time() - self._cache_interval:
            if self.logger:
                self.logger.log(f"Deleting expired cached data from {filename}")
            os.remove(filename)
            return False

        if self.logger:
            self.logger.log(f"Loading cached data from {filename}")

        return filename

    def _save_cache(self, df):
        """
        Save the given DataFrame to the cache.

        Args:
            df (DataFrame): The DataFrame to save to the cache.
        """

        filename = self._get_cache_path(
            include_additional_filters=self._additional_filters is not None
            and len(self._additional_filters) > 0
        )

        df.to_csv(filename, index=False)

    def execute(self):
        """
        Execute the SQL query and return the result.

        Returns:
            DataFrame: The result of the SQL query.
        """

        if cached := self._cached() or self._cached(include_additional_filters=True):
            return pd.read_csv(cached)

        if self.logger:
            self.logger.log(
                f"Loading the table {self.config.table} "
                f"using dialect {self.config.dialect}"
            )

        # Run a SQL query to get all the results
        query = self._build_query()

        # Get the result of the query
        result = pd.read_sql(query, self._connection)

        # Save the result to the cache
        self._save_cache(result)

        # Return the result
        return result

    @cached_property
    def rows_count(self):
        """
        Return the number of rows in the SQL table.

        Returns:
            int: The number of rows in the SQL table.
        """

        if self._rows_count is not None:
            return self._rows_count

        if self.logger:
            self.logger.log(
                "Getting the number of rows in the table "
                f"{self.config.table} using dialect "
                f"{self.config.dialect}"
            )

        # Run a SQL query to get the number of rows
        query = select(text("COUNT(*)")).select_from(text(self.cs_table_name))

        # Return the number of rows
        self._rows_count = self._connection.execute(query).fetchone()[0]
        return self._rows_count

    @cached_property
    def columns_count(self):
        """
        Return the number of columns in the SQL table.

        Returns:
            int: The number of columns in the SQL table.
        """

        if self._columns_count is not None:
            return self._columns_count

        if self.logger:
            self.logger.log(
                "Getting the number of columns in the table "
                f"{self.config.table} using dialect "
                f"{self.config.dialect}"
            )

        self._columns_count = len(self.head().columns)
        return self._columns_count

    def _get_column_hash(self, include_additional_filters: bool = False):
        """
        Return the hash of the SQL table columns.

        Args:
            include_additional_filters (bool, optional): Whether to include the
                additional filters in the hash. Defaults to False.

        Returns:
            str: The hash of the SQL table columns.
        """

        # Return the hash of the columns and the where clause
        columns_str = "".join(self.head().columns)
        if (
            self.config.where
            or include_additional_filters
            and self._additional_filters is not None
        ):
            columns_str += "WHERE"
        if self.config.where:
            # where clause is a list of lists
            for condition in self.config.where:
                columns_str += f"{condition[0]} {condition[1]} {condition[2]}"
        if include_additional_filters and self._additional_filters:
            for condition in self._additional_filters:
                columns_str += f"{condition[0]} {condition[1]} {condition[2]}"

        hash_object = hashlib.sha256(columns_str.encode())
        return hash_object.hexdigest()

    @cached_property
    def column_hash(self):
        """
        Return the hash of the SQL table columns.

        Returns:
            str: The hash of the SQL table columns.
        """
        return self._get_column_hash()

    @property
    def fallback_name(self):
        return self.config.table

    @property
    def pandas_df(self):
        return self.execute()

    def equals(self, other):
        if isinstance(other, self.__class__):
            return (
                self.config.dialect,
                self.config.driver,
                self.config.host,
                self.config.port,
            ) == (
                other.config.dialect,
                other.config.driver,
                other.config.host,
                other.config.port,
            )
        return False

    def _is_sql_query_safe(self, query: str):
        infected_keywords = [
            r"\bINSERT\b",
            r"\bUPDATE\b",
            r"\bDELETE\b",
            r"\bDROP\b",
            r"\bEXEC\b",
            r"\bALTER\b",
            r"\bCREATE\b",
        ]

        return not any(
            re.search(keyword, query, re.IGNORECASE) for keyword in infected_keywords
        )

    def execute_direct_sql_query(self, sql_query):
        if not self._is_sql_query_safe(sql_query):
            raise MaliciousQueryError("Malicious query is generated in code")

        return pd.read_sql(sql_query, self._connection)

    @property
    def cs_table_name(self):
        return self.config.table

    @property
    def type(self):
        return self.config.dialect


class SqliteConnector(SQLConnector):
    """
    Sqlite connector are used to connect to Sqlite databases.
    """

    def __init__(
        self,
        config: Union[SqliteConnectorConfig, dict],
        **kwargs,
    ):
        """
        Initialize the Sqlite connector with the given configuration.

        Args:
            config (ConnectorConfig) : The configuration for the MySQL connector.
        """
        config["dialect"] = "sqlite"
        if isinstance(config, dict):
            sqlite_env_vars = {"database": "SQLITE_DB_PATH", "table": "TABLENAME"}
            config = self._populate_config_from_env(config, sqlite_env_vars)

        super().__init__(config, **kwargs)

    def _load_connector_config(self, config: Union[BaseConnectorConfig, dict]):
        """
        Loads passed Configuration to object

        Args:
            config (BaseConnectorConfig): Construct config in structure

        Returns:
            config: BaseConenctorConfig
        """
        return SqliteConnectorConfig(**config)

    def _init_connection(self, config: SqliteConnectorConfig):
        """
        Initialize Database Connection

        Args:
            config (SQLConnectorConfig): Configurations to load database

        """
        self._engine = create_engine(f"{config.dialect}:///{config.database}")
        self._connection = self._engine.connect()

    def __del__(self):
        """
        Close the connection to the SQL database.
        """
        if self._connection:
            self._connection.close()

    @cache
    def head(self, n: int = 5) -> pd.DataFrame:
        """
        Return the head of the data source that the connector is connected to.
        This information is passed to the LLM to provide the schema of the data source.

        Returns:
            DataFrame: The head of the data source.
        """

        if self.logger:
            self.logger.log(
                f"Getting head of {self.config.table} "
                f"using dialect {self.config.dialect}"
            )

        # Run a SQL query to get all the columns names and 5 random rows
        query = self._build_query(limit=n, order="RANDOM()")

        # Return the head of the data source
        return pd.read_sql(query, self._connection)

    @property
    def cs_table_name(self):
        return f'"{self.config.table}"'

    def __repr__(self):
        """
        Return the string representation of the SQL connector.

        Returns:
            str: The string representation of the SQL connector.
        """
        return (
            f"<{self.__class__.__name__} dialect={self.config.dialect} "
            f"database={self.config.database} "
            f"table={self.config.table}>"
        )

    def equals(self, other):
        if isinstance(other, self.__class__):
            print(self.config.database)
            print(other.config.database)
            return (
                self.config.dialect,
                self.config.driver,
                self.config.database,
            ) == (
                other.config.dialect,
                other.config.driver,
                other.config.database,
            )
        return False


class MySQLConnector(SQLConnector):
    """
    MySQL connectors are used to connect to MySQL databases.
    """

    def __init__(
        self,
        config: Union[SQLConnectorConfig, dict],
        **kwargs,
    ):
        """
        Initialize the MySQL connector with the given configuration.

        Args:
            config (ConnectorConfig): The configuration for the MySQL connector.
        """
        config["dialect"] = "mysql"
        config["driver"] = "pymysql"

        if isinstance(config, dict):
            mysql_env_vars = {
                "host": "MYSQL_HOST",
                "port": "MYSQL_PORT",
                "database": "MYSQL_DATABASE",
                "username": "MYSQL_USERNAME",
                "password": "MYSQL_PASSWORD",
            }
            config = self._populate_config_from_env(config, mysql_env_vars)

        super().__init__(config, **kwargs)


class PostgreSQLConnector(SQLConnector):
    """
    PostgreSQL connectors are used to connect to PostgreSQL databases.
    """

    def __init__(
        self,
        config: Union[SQLConnectorConfig, dict],
        **kwargs,
    ):
        """
        Initialize the PostgreSQL connector with the given configuration.

        Args:
            config (ConnectorConfig): The configuration for the PostgreSQL connector.
        """
        if "dialect" not in config:
            config["dialect"] = "postgresql"

        config["driver"] = "psycopg2"

        if isinstance(config, dict):
            postgresql_env_vars = {
                "host": "POSTGRESQL_HOST",
                "port": "POSTGRESQL_PORT",
                "database": "POSTGRESQL_DATABASE",
                "username": "POSTGRESQL_USERNAME",
                "password": "POSTGRESQL_PASSWORD",
            }
            config = self._populate_config_from_env(config, postgresql_env_vars)

        super().__init__(config, **kwargs)

    @cache
    def head(self, n: int = 5) -> pd.DataFrame:
        """
        Return the head of the data source that the connector is connected to.
        This information is passed to the LLM to provide the schema of the data source.

        Returns:
            DataFrame: The head of the data source.
        """

        if self.logger:
            self.logger.log(
                f"Getting head of {self.config.table} "
                f"using dialect {self.config.dialect}"
            )

        # Run a SQL query to get all the columns names and 5 random rows
        query = self._build_query(limit=n, order="RANDOM()")

        # Return the head of the data source
        return pd.read_sql(query, self._connection)

    @property
    def cs_table_name(self):
        return f'"{self.config.table}"'

    def execute_direct_sql_query(self, sql_query):
        sql_query = sqlglot.transpile(sql_query, read="mysql", write="postgres")[0]
        return super().execute_direct_sql_query(sql_query)


class OracleConnector(SQLConnector):
    """
    Oracle connectors are used to connect to Oracle databases.
    """

    def __init__(
        self,
        config: Union[SQLConnectorConfig, dict],
        **kwargs,
    ):
        """
        Initialize the Oracle connector with the given configuration.

        Args:
            config (ConnectorConfig): The configuration for the Oracle connector.
        """
        config["dialect"] = "oracle"
        config["driver"] = "cx_oracle"

        if isinstance(config, dict):
            oracle_env_vars = {
                "host": "ORACLE_HOST",
                "port": "ORACLE_PORT",
                "database": "ORACLE_DATABASE",
                "username": "ORACLE_USERNAME",
                "password": "ORACLE_PASSWORD",
            }
            config = self._populate_config_from_env(config, oracle_env_vars)

        super().__init__(config, **kwargs)

    @cache
    def head(self, n: int = 5) -> pd.DataFrame:
        """
        Return the head of the data source that the connector is connected to.
        This information is passed to the LLM to provide the schema of the data source.

        Returns:
            DataFrame: The head of the data source.
        """

        if self.logger:
            self.logger.log(
                f"Getting head of {self.config.table} "
                f"using dialect {self.config.dialect}"
            )

        # Run a SQL query to get all the columns names and 5 random rows
        query = self._build_query(limit=n, order="dbms_random.value")

        # Return the head of the data source
        return pd.read_sql(query, self._connection)

    @property
    def cs_table_name(self):
        return f'"{self.config.table}"'



import os

from pandasai.exceptions import InvalidWorkspacePathError


def find_project_root(filename=None):
    """
    Check if Custom workspace path provide use that otherwise iterate to
    find project root
    """
    if "PANDASAI_WORKSPACE" in os.environ:
        workspace_path = os.environ["PANDASAI_WORKSPACE"]
        if (
            workspace_path
            and os.path.exists(workspace_path)
            and os.path.isdir(workspace_path)
        ):
            return workspace_path
        raise InvalidWorkspacePathError(
            "PANDASAI_WORKSPACE does not point to a valid directory"
        )

    # Get the path of the file that is be
    # ing executed
    current_file_path = os.path.abspath(os.getcwd())

    # Navigate back until we either find a $filename file or there is no parent
    # directory left.
    root_folder = current_file_path
    while True:
        # Custom way to identify the project root folder
        if filename is not None:
            env_file_path = os.path.join(root_folder, filename)
            if os.path.isfile(env_file_path):
                break

        # Most common ways to identify a project root folder
        if (
            os.path.isfile(os.path.join(root_folder, "pyproject.toml"))
            or os.path.isfile(os.path.join(root_folder, "setup.py"))
            or os.path.isfile(os.path.join(root_folder, "requirements.txt"))
            or os.path.isfile(os.path.join(root_folder, "pandasai.json"))
        ):
            break

        parent_folder = os.path.dirname(root_folder)
        if parent_folder == root_folder:
            # if project root is not found return cwd
            return os.getcwd()

        root_folder = parent_folder

    return root_folder


def find_closest(filename):
    return os.path.join(find_project_root(filename), filename)


def create_directory(path):
    if not os.path.exists(path):
        # Create the directory
        try:
            os.makedirs(path)
        except OSError as e:
            raise OSError(f"Error creating directory: {e}") from e



# -*- coding: utf-8 -*-
"""
PandasAI is a wrapper around a LLM to make dataframes conversational
"""

from pandasai.smart_dataframe import SmartDataframe
from pandasai.smart_datalake import SmartDatalake

from .agent import Agent
from .engine import set_pd_engine
from .helpers.cache import Cache
from .skills import skill


def clear_cache(filename: str = None):
    """Clear the cache"""
    cache = Cache(filename) if filename else Cache()

    cache.clear()


__all__ = [
    "Agent",
    "clear_cache",
    "skill",
    "set_pd_engine",
    "pandas",
    "SmartDataframe",
    "SmartDatalake",
]



from dataclasses import dataclass


@dataclass
class ErrorCorrectionPipelineInput:
    code: str
    exception: Exception

    def __init__(self, code: str, exception: Exception):
        self.code = code
        self.exception = exception



from typing import Any

from pandasai.pipelines.logic_unit_output import LogicUnitOutput

from ...helpers.logger import Logger
from ..base_logic_unit import BaseLogicUnit
from ..pipeline_context import PipelineContext


class CacheLookup(BaseLogicUnit):
    """
    Cache Lookup of Code Stage
    """

    pass

    def execute(self, input: Any, **kwargs) -> Any:
        """
        This method will return output according to
        Implementation.

        :param input: Your input data.
        :param kwargs: A dictionary of keyword arguments.
            - 'logger' (any): The logger for logging.
            - 'config' (Config): Global configurations for the test
            - 'context' (any): The execution context.

        :return: The result of the execution.
        """
        pipeline_context: PipelineContext = kwargs.get("context")
        logger: Logger = kwargs.get("logger")
        if (
            pipeline_context.config.enable_cache
            and pipeline_context.cache
            and pipeline_context.cache.get(
                pipeline_context.cache.get_cache_key(pipeline_context)
            )
        ):
            logger.log("Using cached response")

            code = pipeline_context.cache.get(
                pipeline_context.cache.get_cache_key(pipeline_context)
            )

            pipeline_context.add("found_in_cache", True)

            return LogicUnitOutput(code, True, "Cache Hit")



"""
Airtable connectors are used to connect airtable records.
"""

import hashlib
import os
import time
from functools import cache, cached_property
from typing import Optional, Union

import requests

import pandasai.pandas as pd

from ..exceptions import InvalidRequestError
from ..helpers.path import find_project_root
from .base import BaseConnector, BaseConnectorConfig


class AirtableConnectorConfig(BaseConnectorConfig):
    """
    Connecter configuration for Airtable data.
    """

    api_key: str
    base_id: str
    database: str = "airtable_data"


class AirtableConnector(BaseConnector):
    """
    Airtable connector to retrieving record data.
    """

    _rows_count: int = None
    _columns_count: int = None
    _instance = None

    def __init__(
        self,
        config: Optional[Union[AirtableConnectorConfig, dict]] = None,
        cache_interval: int = 600,
        **kwargs,
    ):
        if isinstance(config, dict):
            if "token" in config and "base_id" in config and "table" in config:
                config = AirtableConnectorConfig(**config)
            else:
                raise KeyError(
                    "Please specify all token, table, base_id properly in config ."
                )

        elif not config:
            airtable_env_vars = {
                "token": "AIRTABLE_API_TOKEN",
                "base_id": "AIRTABLE_BASE_ID",
                "table": "AIRTABLE_TABLE_NAME",
            }
            config = AirtableConnectorConfig(
                **self._populate_config_from_env(config, airtable_env_vars)
            )

        self._root_url: str = "https://api.airtable.com/v0/"
        self._cache_interval = cache_interval

        super().__init__(config, **kwargs)

    def _init_connection(self, config: BaseConnectorConfig):
        """
        make connection to database
        """
        config = config.dict()
        url = f"{self._root_url}{config['base_id']}/{config['table']}"
        response = requests.head(
            url=url, headers={"Authorization": f"Bearer {config['token']}"}
        )
        if response.status_code == 200:
            self.logger.log(
                """
                Connected to Airtable.
            """
            )
        else:
            raise InvalidRequestError(
                f"""Failed to connect to Airtable. 
                    Status code: {response.status_code}, 
                    message: {response.text}"""
            )

    def _get_cache_path(self, include_additional_filters: bool = False):
        """
        Return the path of the cache file.

        Returns :
            str : The path of the cache file.
        """
        cache_dir = os.path.join(os.getcwd(), "")
        try:
            cache_dir = os.path.join((find_project_root()), "cache")
        except ValueError:
            cache_dir = os.path.join(os.getcwd(), "cache")
        return os.path.join(cache_dir, f"{self.config.table}_data.parquet")

    def _cached(self, include_additional_filters: bool = False):
        """
        Returns the cached Airtable data if it exists and
        is not older than the cache interval.

        Returns :
            DataFrame | None : The cached data if
                it exists and is not older than the cache
                interval, None otherwise.
        """
        cache_path = self._get_cache_path(include_additional_filters)
        if not os.path.exists(cache_path):
            return None

        # If the file is older than 1 day , delete it.
        if os.path.getmtime(cache_path) < time.time() - self._cache_interval:
            if self.logger:
                self.logger.log(f"Deleting expired cached data from {cache_path}")
            os.remove(cache_path)
            return None

        if self.logger:
            self.logger.log(f"Loading cached data from {cache_path}")

        return cache_path

    def _save_cache(self, df):
        """
        Save the given DataFrame to the cache.

        Args:
            df (DataFrame): The DataFrame to save to the cache.
        """
        filename = self._get_cache_path(
            include_additional_filters=self._additional_filters is not None
            and len(self._additional_filters) > 0
        )
        df.to_parquet(filename)

    @property
    def fallback_name(self):
        """
        Returns the fallback table name of the connector.

        Returns :
            str : The fallback table name of the connector.
        """
        return self.config.table

    def execute(self) -> pd.DataFrame:
        """
        Execute the connector and return the result.

        Returns:
            pd.DataFrame: The result of the connector.
        """
        if cached := self._cached() or self._cached(include_additional_filters=True):
            return pd.read_parquet(cached)

        if isinstance(self._instance, pd.DataFrame):
            return self._instance
        else:
            self._instance = self._fetch_data()

        return self._instance

    def _build_formula(self):
        """
        Build Airtable query formula for filtering.
        """

        condition_strings = []
        if self.config.where is not None:
            for i in self.config.where:
                filter_query = f"{i[0]}{i[1]}'{i[2]}'"
                condition_strings.append(filter_query)
        return f'AND({",".join(condition_strings)})'

    def _request_api(self, params):
        url = f"{self._root_url}{self.config.base_id}/{self.config.table}"
        return requests.get(
            url=url,
            headers={"Authorization": f"Bearer {self.config.api_key}"},
            params=params,
        )

    def _fetch_data(self):
        """
        Fetches data from the Airtable server via API and converts it to a DataFrame.
        """

        params = {"pageSize": 100, "offset": "0"}

        if self.config.where is not None:
            params["filterByFormula"] = self._build_formula()

        data = []
        while True:
            response = self._request_api(params=params)

            if response.status_code != 200:
                raise InvalidRequestError(
                    f"Failed to connect to Airtable. "
                    f"Status code: {response.status_code}, "
                    f"message: {response.text}"
                )

            res = response.json()
            records = res.get("records", [])
            data.extend({"id": record["id"], **record["fields"]} for record in records)

            if len(records) < 100 or "offset" not in res:
                break

            params["offset"] = res["offset"]

        return pd.DataFrame(data)

    @cache
    def head(self, n: int = 5) -> pd.DataFrame:
        """
        Return the head of the table that
          the connector is connected to.

        Returns :
            DatFrameType: The head of the data source
                 that the connector is connected to .
        """
        data = self._request_api(params={"maxRecords": n})
        return pd.DataFrame(
            [
                {"id": record["id"], **record["fields"]}
                for record in data.json()["records"]
            ]
        )

    @cached_property
    def rows_count(self):
        """
        Return the number of rows in the data source that the connector is
        connected to.

        Returns:
            int: The number of rows in the data source that the connector is
            connected to.
        """
        if self._rows_count is not None:
            return self._rows_count
        data = self.execute()
        self._rows_count = len(data)
        return self._rows_count

    @cached_property
    def columns_count(self):
        """
        Return the number of columns in the data source that the connector is
        connected to.

        Returns:
            int: The number of columns in the data source that the connector is
            connected to.
        """
        if self._columns_count is not None:
            return self._columns_count
        data = self.head()
        self._columns_count = len(data.columns)
        return self._columns_count

    @property
    def column_hash(self):
        """
        Return the hash code that is unique to the columns of the data source
        that the connector is connected to.

        Returns:
            int: The hash code that is unique to the columns of the data source
            that the connector is connected to.
        """
        if not isinstance(self._instance, pd.DataFrame):
            self._instance = self.execute()
        columns_str = "|".join(self._instance.columns)
        columns_str += f"WHERE{self._build_formula()}"
        return hashlib.sha256(columns_str.encode("utf-8")).hexdigest()



from typing import Any

from pandasai.pipelines.logic_unit_output import LogicUnitOutput

from ..base_logic_unit import BaseLogicUnit
from ..pipeline_context import PipelineContext


class CachePopulation(BaseLogicUnit):
    """
    Cache Population Stage
    """

    pass

    def execute(self, input: Any, **kwargs) -> Any:
        """
        This method will return output according to
        Implementation.

        :param input: Your input data.
        :param kwargs: A dictionary of keyword arguments.
            - 'logger' (any): The logger for logging.
            - 'config' (Config): Global configurations for the test
            - 'context' (any): The execution context.

        :return: The result of the execution.
        """
        pipeline_context: PipelineContext = kwargs.get("context")

        code = input

        if pipeline_context.config.enable_cache and pipeline_context.cache:
            pipeline_context.cache.set(
                pipeline_context.cache.get_cache_key(pipeline_context), code
            )

        return LogicUnitOutput(
            code,
            True,
            "Prompt Cached Successfully"
            if pipeline_context.config.enable_cache
            else "Caching disabled",
        )



"""PandasAI's custom exceptions.

This module contains the implementation of Custom Exceptions.

"""

from pandasai.constants import PANDASBI_SETUP_MESSAGE


class InvalidRequestError(Exception):
    """
    Raised when the request is not successful.

    Args :
        Exception (Exception): InvalidRequestError
    """


class APIKeyNotFoundError(Exception):
    """
    Raised when the API key is not defined/declared.

    Args:
        Exception (Exception): APIKeyNotFoundError
    """


class LLMNotFoundError(Exception):
    """
    Raised when the LLM is not provided.

    Args:
        Exception (Exception): LLMNotFoundError
    """


class NoCodeFoundError(Exception):
    """
    Raised when no code is found in the response.

    Args:
        Exception (Exception): NoCodeFoundError
    """


class NoResultFoundError(Exception):
    """
    Raised when no result is found in the response.

    Args:
        Exception (Exception): NoResultFoundError
    """


class MethodNotImplementedError(Exception):
    """
    Raised when a method is not implemented.

    Args:
        Exception (Exception): MethodNotImplementedError
    """


class UnsupportedModelError(Exception):
    """
    Raised when an unsupported model is used.

    Args:
        model_name (str): The name of the unsupported model.
        Exception (Exception): UnsupportedModelError
    """

    def __init__(self, model_name):
        self.model = model_name
        super().__init__(
            f"Unsupported model: The model '{model_name}' doesn't exist "
            f"or is not supported yet."
        )


class MissingModelError(Exception):
    """
    Raised when deployment name is not passed to azure as it's a required parameter

    Args:
    Exception (Exception): MissingModelError
    """


class LLMResponseHTTPError(Exception):
    """
    Raised when a remote LLM service responses with error HTTP code.

    Args:
        Exception (Exception): LLMResponseHTTPError
    """

    def __init__(self, status_code, error_msg=None):
        self.status_code = status_code
        self.error_msg = error_msg

        super().__init__(
            f"The remote server has responded with an error HTTP "
            f"code: {status_code}; {error_msg or ''}"
        )


class BadImportError(Exception):
    """
    Raised when a library not in the whitelist is imported.

    Args:
        Exception (Exception): BadImportError
    """

    def __init__(self, library_name):
        """
        __init__ method of BadImportError Class

        Args:
            library_name (str): Name of the library that is not in the whitelist.
        """
        self.library_name = library_name
        super().__init__(
            f"Generated code includes import of {library_name} which"
            " is not in whitelist."
        )


class TemplateFileNotFoundError(FileNotFoundError):
    """
    Raised when a template file cannot be found.
    """

    def __init__(self, template_path, prompt_name="Unknown"):
        """
        __init__ method of TemplateFileNotFoundError Class

        Args:
            template_path (str): Path for template file.
            prompt_name (str): Prompt name. Defaults to "Unknown".
        """
        self.template_path = template_path
        super().__init__(
            f"Unable to find a file with template at '{template_path}' "
            f"for '{prompt_name}' prompt."
        )


class UnSupportedLogicUnit(Exception):
    """
    Raised when unsupported logic unit is added in the pipeline
    Args:
        Exception (Exception): UnSupportedLogicUnit
    """


class InvalidWorkspacePathError(Exception):
    """
    Raised when the environment variable of workspace exist but path is invalid

    Args:
        Exception (Exception): InvalidWorkspacePathError
    """


class InvalidConfigError(Exception):
    """
    Raised when config value is not applicable
    Args:
        Exception (Exception): InvalidConfigError
    """


class MaliciousQueryError(Exception):
    """
    Raise error if malicious query is generated
    Args:
        Exception (Exception): MaliciousQueryError
    """


class InvalidLLMOutputType(Exception):
    """
    Raise error if the output type is invalid
    Args:
        Exception (Exception): InvalidLLMOutputType
    """


class InvalidOutputValueMismatch(Exception):
    """
    Raise error if the output value doesn't match with type
    Args:
        Exception (Exception): InvalidOutputValueMismatch
    """


class ExecuteSQLQueryNotUsed(Exception):
    """
    Raise error if Execute SQL Query is not used
    Args:
        Exception (Exception): ExecuteSQLQueryNotUsed
    """


class PipelineConcatenationError(Exception):
    """
    Raise error if vector store is not found
    Args:
        Exception (Exception): Concatenating wrong pipelines
    """


class MissingVectorStoreError(Exception):
    """
    Raise error if vector store is not found
    Args:
        Exception (Exception): MissingVectorStoreError
    """


class PandasAIApiKeyError(Exception):
    """
    Raise error if api key is not found for remote vectorstore and llm
    Args:
        Exception (Exception): PandasAIApiKeyError
    """

    def __init__(self):
        message = PANDASBI_SETUP_MESSAGE
        super().__init__(message)


class PandasAIApiCallError(Exception):
    """
    Raise error if exception in API request fails
    Args:
        Exception (Exception): PandasAIApiCallError
    """


class PandasConnectorTableNotFound(Exception):
    """
    Raise error if exception in API request fails
    Args:
        Exception (Exception): PandasConnectorTableNotFound
    """


class InvalidTrainJson(Exception):
    """
    Raise error if train json is not correct
    Args:
        Exception (Exception): Invalid train json
    """


class InvalidSchemaJson(Exception):
    """
    Raise error if schema json is not correct
    Args:
        Exception (Exception): Invalid json schema
    """



"""Google VertexAI

This module is to run the Google VertexAI LLM.
To read more on VertexAI:
https://cloud.google.com/vertex-ai/docs/generative-ai/learn/generative-ai-studio.

Example:
    Use below example to call Google VertexAI

    >>> from pandasai.llm.google_palm import GoogleVertexAI

"""
from typing import Optional

from pandasai.helpers.memory import Memory

from ..exceptions import UnsupportedModelError
from ..helpers.optional import import_dependency
from .base import BaseGoogle


class GoogleVertexAI(BaseGoogle):
    """Google Palm Vertexai LLM
    BaseGoogle class is extended for Google Palm model using Vertexai.
    The default model support at the moment is text-bison-001.
    However, user can choose to use code-bison-001 too.
    """

    _supported_code_models = [
        "code-bison",
        "code-bison-32k",
        "code-bison-32k@002",
        "code-bison@001",
        "code-bison@002",
    ]
    _supported_text_models = [
        "text-bison",
        "text-bison-32k",
        "text-bison-32k@002",
        "text-bison@001",
        "text-bison@002",
        "text-unicorn@001",
    ]
    _supported_generative_models = [
        "gemini-pro",
    ]
    _supported_code_chat_models = ["codechat-bison@001", "codechat-bison@002"]

    def __init__(
        self, project_id: str, location: str, model: Optional[str] = None, **kwargs
    ):
        """
        A init class to implement the Google Vertexai Models

        Args:
            project_id (str): GCP project
            location (str): GCP project Location
            model Optional (str): Model to use Default to text-bison@001
            **kwargs: Arguments to control the Model Parameters
        """

        self.model = model or "text-bison@001"

        self._configure(project_id, location)
        self.project_id = project_id
        self.location = location
        self._set_params(**kwargs)

    def _configure(self, project_id: str, location: str):
        """
        Configure Google VertexAi. Set value `self.vertexai` attribute.

        Args:
            project_id (str): GCP Project.
            location (str): Location of Project.

        Returns:
            None.

        """

        err_msg = "Install google-cloud-aiplatform for Google Vertexai"
        vertexai = import_dependency("vertexai", extra=err_msg)
        vertexai.init(project=project_id, location=location)
        self.vertexai = vertexai

    def _valid_params(self):
        """Returns if the Parameters are valid or Not"""
        return super()._valid_params() + ["model"]

    def _validate(self):
        """
        A method to Validate the Model

        """

        super()._validate()

        if not self.model:
            raise ValueError("model is required.")

    def _generate_text(self, prompt: str, memory: Optional[Memory] = None) -> str:
        """
        Generates text for prompt.

        Args:
            prompt (str): A string representation of the prompt.

        Returns:
            str: LLM response.

        """
        self._validate()

        updated_prompt = self.prepend_system_prompt(prompt, memory)

        self.last_prompt = updated_prompt

        if self.model in self._supported_code_models:
            from vertexai.preview.language_models import CodeGenerationModel

            code_generation = CodeGenerationModel.from_pretrained(self.model)

            completion = code_generation.predict(
                prefix=prompt,
                temperature=self.temperature,
                max_output_tokens=self.max_output_tokens,
            )
        elif self.model in self._supported_text_models:
            from vertexai.preview.language_models import TextGenerationModel

            text_generation = TextGenerationModel.from_pretrained(self.model)

            completion = text_generation.predict(
                prompt=updated_prompt,
                temperature=self.temperature,
                top_p=self.top_p,
                top_k=self.top_k,
                max_output_tokens=self.max_output_tokens,
            )
        elif self.model in self._supported_generative_models:
            from vertexai.preview.generative_models import GenerativeModel

            model = GenerativeModel(self.model)

            responses = model.generate_content(
                [updated_prompt],
                generation_config={
                    "max_output_tokens": self.max_output_tokens,
                    "temperature": self.temperature,
                    "top_p": self.top_p,
                    "top_k": self.top_k,
                },
            )

            completion = responses.candidates[0].content.parts[0]
        elif self.model in self._supported_code_chat_models:
            from vertexai.language_models import ChatMessage, CodeChatModel

            code_chat_model = CodeChatModel.from_pretrained(self.model)
            messages = []

            for message in memory.all():
                if message["is_user"]:
                    messages.append(
                        ChatMessage(author="user", content=message["message"])
                    )
                else:
                    messages.append(
                        ChatMessage(author="model", content=message["message"])
                    )
            chat = code_chat_model.start_chat(
                context=memory.get_system_prompt(), message_history=messages
            )

            response = chat.send_message(prompt)
            return response.text

        else:
            raise UnsupportedModelError(self.model)

        return completion.text

    @property
    def type(self) -> str:
        return "google-vertexai"



"""Fake LLM"""

from typing import Optional

from pandasai.pipelines.pipeline_context import PipelineContext

from ..prompts.base import BasePrompt
from .base import LLM


class FakeLLM(LLM):
    """Fake LLM"""

    _output: str = """result = { 'type': 'string', 'value': "Hello World" }"""

    def __init__(self, output: Optional[str] = None):
        if output is not None:
            self._output = output

    def call(self, instruction: BasePrompt, context: PipelineContext = None) -> str:
        self.last_prompt = instruction.to_string()
        return self._output

    @property
    def type(self) -> str:
        return "fake"



import os
import uuid
from typing import Callable, Iterable, List, Optional, Union

import chromadb
from chromadb import config
from chromadb.utils import embedding_functions

from pandasai.helpers.logger import Logger
from pandasai.helpers.path import find_project_root
from pandasai.vectorstores.vectorstore import VectorStore

DEFAULT_EMBEDDING_FUNCTION = embedding_functions.DefaultEmbeddingFunction()


class ChromaDB(VectorStore):
    """
    Implementation of ChromeDB vector store
    """

    _logger: Logger

    def __init__(
        self,
        collection_name: str = "pandasai",
        embedding_function: Optional[Callable[[List[str]], List[float]]] = None,
        persist_path: Optional[str] = None,
        client_settings: Optional[config.Settings] = None,
        max_samples: int = 1,
        similary_threshold: int = 1.5,
        logger: Optional[Logger] = None,
    ) -> None:
        self._logger = logger or Logger()
        self._max_samples = max_samples
        self._similarity_threshold = similary_threshold

        # Initialize Chromadb Client
        # initialize from client settings if exists
        if client_settings:
            client_settings.persist_directory = (
                persist_path or client_settings.persist_directory
            )
            _client_settings = client_settings

        # use persist path if exists
        elif persist_path:
            _client_settings = config.Settings(
                is_persistent=True, anonymized_telemetry=False
            )
            _client_settings.persist_directory = persist_path
        # else use root as default path
        else:
            _client_settings = config.Settings(
                is_persistent=True, anonymized_telemetry=False
            )
            _client_settings.persist_directory = os.path.join(
                find_project_root(), "chromadb"
            )

        self._client_settings = _client_settings
        self._client = chromadb.Client(_client_settings)
        self._persist_directory = _client_settings.persist_directory

        self._logger.log(f"Persisting Agent Training data in {self._persist_directory}")

        self._embedding_function = embedding_function or DEFAULT_EMBEDDING_FUNCTION

        self._qa_collection = self._client.get_or_create_collection(
            name=f"{collection_name}-qa", embedding_function=self._embedding_function
        )

        self._docs_collection = self._client.get_or_create_collection(
            name=f"{collection_name}-docs", embedding_function=self._embedding_function
        )

        self._logger.log(f"Successfully initialized collection {collection_name}")

    def add_question_answer(
        self,
        queries: Iterable[str],
        codes: Iterable[str],
        ids: Optional[Iterable[str]] = None,
        metadatas: Optional[List[dict]] = None,
    ) -> List[str]:
        """
        Add question and answer(code) to the training set
        Args:
            query: string of question
            code: str
            ids: Optional Iterable of ids associated with the texts.
            metadatas: Optional list of metadatas associated with the texts.
            kwargs: vectorstore specific parameters
        Returns:
            List of ids from adding the texts into the vectorstore.
        """
        if len(queries) != len(codes):
            raise ValueError(
                f"Queries and codes dimension doesn't match {len(queries)} != {len(codes)}"
            )

        if ids is None:
            ids = [f"{str(uuid.uuid4())}-qa" for _ in queries]
        qa_str = [self._format_qa(query, code) for query, code in zip(queries, codes)]

        self._qa_collection.add(
            documents=qa_str,
            metadatas=metadatas,
            ids=ids,
        )

    def add_docs(
        self,
        docs: Iterable[str],
        ids: Optional[Iterable[str]] = None,
        metadatas: Optional[List[dict]] = None,
    ) -> List[str]:
        """
        Add docs to the training set
        Args:
            docs: Iterable of strings to add to the vectorstore.
            ids: Optional Iterable of ids associated with the texts.
            metadatas: Optional list of metadatas associated with the texts.
            kwargs: vectorstore specific parameters

        Returns:
            List of ids from adding the texts into the vectorstore.
        """
        if ids is None:
            ids = [f"{str(uuid.uuid4())}-docs" for _ in docs]
        self._docs_collection.add(
            documents=docs,
            metadatas=metadatas,
            ids=ids,
        )

    def update_question_answer(
        self,
        ids: Iterable[str],
        queries: Iterable[str],
        codes: Iterable[str],
        metadatas: Optional[List[dict]] = None,
    ) -> List[str]:
        """
        Update question and answer(code) to the training set
        Args:
            ids: Iterable of ids associated with the texts.
            queries: string of question
            codes: str
            metadatas: Optional list of metadatas associated with the texts.
            kwargs: vectorstore specific parameters
        Returns:
            List of ids from updating the texts into the vectorstore.
        """
        if len(queries) != len(codes):
            raise ValueError(
                f"Queries and codes dimension doesn't match {len(queries)} != {len(codes)}"
            )

        qa_str = [self._format_qa(query, code) for query, code in zip(queries, codes)]
        self._qa_collection.update(
            documents=qa_str,
            metadatas=metadatas,
            ids=ids,
        )

    def update_docs(
        self,
        ids: Iterable[str],
        docs: Iterable[str],
        metadatas: Optional[List[dict]] = None,
    ) -> List[str]:
        """
        Update docs to the training set
        Args:
            ids: Iterable of ids associated with the texts.
            docs: Iterable of strings to update to the vectorstore.
            metadatas: Optional list of metadatas associated with the texts.
            kwargs: vectorstore specific parameters

        Returns:
            List of ids from adding the texts into the vectorstore.
        """
        self._docs_collection.update(
            documents=docs,
            metadatas=metadatas,
            ids=ids,
        )

    def delete_question_and_answers(
        self, ids: Optional[List[str]] = None
    ) -> Optional[bool]:
        """
        Delete by vector ID to delete question and answers
        Args:
            ids: List of ids to delete

        Returns:
            Optional[bool]: True if deletion is successful,
            False otherwise
        """
        self._qa_collection.delete(ids=ids)
        return True

    def delete_docs(self, ids: Optional[List[str]] = None) -> Optional[bool]:
        """
        Delete by vector ID to delete docs
        Args:
            ids: List of ids to delete

        Returns:
            Optional[bool]: True if deletion is successful,
            False otherwise
        """
        self._docs_collection.delete(ids=ids)
        return True

    def get_relevant_question_answers(
        self, question: str, k: Union[int, None] = None
    ) -> List[dict]:
        """
        Returns relevant question answers based on search
        """
        k = k or self._max_samples

        relevant_data: chromadb.QueryResult = self._qa_collection.query(
            query_texts=question,
            n_results=k,
            include=["metadatas", "documents", "distances"],
        )

        return self._filter_docs_based_on_distance(
            relevant_data, self._similarity_threshold
        )

    def get_relevant_docs(self, question: str, k: int = None) -> List[dict]:
        """
        Returns relevant documents based search
        """
        k = k or self._max_samples

        relevant_data: chromadb.QueryResult = self._docs_collection.query(
            query_texts=question,
            n_results=k,
            include=["metadatas", "documents", "distances"],
        )

        return self._filter_docs_based_on_distance(
            relevant_data, self._similarity_threshold
        )

    def get_relevant_question_answers_by_id(self, ids: Iterable[str]) -> List[dict]:
        """
        Returns relevant question answers based on ids
        """

        relevant_data: chromadb.QueryResult = self._qa_collection.get(
            ids=ids,
            include=["metadatas", "documents"],
        )

        return relevant_data

    def get_relevant_docs_by_id(self, ids: Iterable[str]) -> List[dict]:
        """
        Returns relevant question answers based on ids
        """

        relevant_data: chromadb.QueryResult = self._docs_collection.get(
            ids=ids,
            include=["metadatas", "documents"],
        )

        return relevant_data

    def get_relevant_qa_documents(self, question: str, k: int = None) -> List[str]:
        """
        Returns relevant question answers documents only
        Args:
            question (_type_): list of documents
        """
        return self.get_relevant_question_answers(question, k)["documents"][0]

    def get_relevant_docs_documents(self, question: str, k: int = None) -> List[str]:
        """
        Returns relevant question answers documents only
        Args:
            question (_type_): list of documents
        """
        return self.get_relevant_docs(question, k)["documents"][0]

    def _filter_docs_based_on_distance(
        self, documents: chromadb.QueryResult, threshold: int
    ) -> List[str]:
        """
        Filter documents based on threshold
        Args:
            documents (List[str]): list of documents in string
            distances (List[float]): list of distances in float
            threshold (int): similarity threshold

        Returns:
            _type_: _description_
        """
        filtered_data = [
            (doc, distance, metadata, ids)
            for doc, distance, metadata, ids in zip(
                documents["documents"][0],
                documents["distances"][0],
                documents["metadatas"][0],
                documents["ids"][0],
            )
            if distance < threshold
        ]

        return {
            key: [[data[i] for data in filtered_data]]
            for i, key in enumerate(["documents", "distances", "metadatas", "ids"])
        }
